from __future__ import annotations

import asyncio
import mimetypes

import httpx

from pywa._helpers import *  # noqa: F403

import pathlib

from pywa._helpers import (
    _media_types_default_filenames,
    _template_header_formats_filename,
    _template_header_formats_default_mime_types,
)
from pywa.types.templates import (
    TemplateBaseComponent,
    _BaseMediaHeaderComponent,
    Carousel,
    _BaseMediaParams,
)
from .types import MessageType

from typing import BinaryIO, Literal, TYPE_CHECKING

from .types.media import Media

if TYPE_CHECKING:
    from pywa_async import WhatsApp


async def resolve_media_param(
    wa: WhatsApp,
    media: str | Media | pathlib.Path | bytes | BinaryIO,
    mime_type: str | None,
    filename: str | None,
    media_type: Literal[
        MessageType.IMAGE, MessageType.VIDEO, MessageType.AUDIO, MessageType.STICKER
    ]
    | None,
    phone_id: str,
) -> tuple[bool, str]:
    """
    Internal method to resolve the ``media`` parameter. Returns a tuple of (``is_url``, ``media_id_or_url``).
    """
    if isinstance(media, Media):
        return False, media.id
    if isinstance(media, (str, pathlib.Path)):
        if str(media).startswith(("https://", "http://")):
            return True, media
        elif str(media).isdigit() and not pathlib.Path(media).is_file():
            return False, media  # assume it's a media ID
    # assume its bytes or a file path
    return False, (
        await wa.upload_media(
            phone_id=phone_id,
            media=media,
            mime_type=mime_type,
            filename=_media_types_default_filenames.get(media_type, filename),
        )
    ).id


async def upload_template_media_components(
    *,
    wa: WhatsApp,
    app_id: int | str | None,
    components: list[TemplateBaseComponent | dict],
) -> None:
    """
    Internal method to upload media components examples in a template.
    """
    not_uploaded = []
    for comp in components:
        if isinstance(comp, _BaseMediaHeaderComponent) and comp._handle is None:
            not_uploaded.append(comp)
        elif isinstance(comp, Carousel):
            for card in comp.cards:
                for cc in card.components:
                    if isinstance(cc, _BaseMediaHeaderComponent) and cc._handle is None:
                        not_uploaded.append(cc)
    if not_uploaded:
        await _run_all_and_cancel_on_exception(
            *[
                _upload_comp(
                    wa=wa,
                    c=c,
                    app_id=app_id,
                )
                for c in not_uploaded
            ]
        )


async def _run_all_and_cancel_on_exception(*coros):
    tasks = [asyncio.create_task(c) for c in coros]

    try:
        return await asyncio.gather(*tasks)
    except Exception as e:
        for task in tasks:
            if not task.done():
                task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)
        raise e


async def _upload_comp(
    *, wa: WhatsApp, c: _BaseMediaHeaderComponent, app_id: int | str | None
) -> None:
    example = c.example
    filename: str | None = None
    mime_type: str | None = None
    raw_bytes: bytes | None = None
    try:
        if isinstance(example, (str, pathlib.Path, Media)):
            if str(example).startswith(("https://", "http://")):  # URL
                res = await httpx.AsyncClient(follow_redirects=True).get(str(example))
                res.raise_for_status()
                raw_bytes = res.content
                mime_type = res.headers.get("Content-Type")
            elif (path := pathlib.Path(example)).is_file():  # File path
                with open(path, "rb") as f:
                    filename = path.name
                    mime_type = mimetypes.guess_type(path)[0]
                    raw_bytes = f.read()
            elif str(example).isdigit() or isinstance(
                example, Media
            ):  # WhatsApp Media ID
                url_res = await wa.get_media_url(
                    media_id=example.id if isinstance(example, Media) else example
                )
                mime_type = url_res.mime_type
                raw_bytes = await wa.download_media(url=url_res.url, in_memory=True)
        elif isinstance(example, bytes):
            raw_bytes = example
        if not raw_bytes:
            raise ValueError(
                f"Invalid media example for component {c.__class__.__name__}: {example}. "
                "It must be a URL, file path, WhatsApp Media, or bytes."
            )
        app_id = resolve_arg(
            wa=wa,
            value=app_id,
            method_arg="app_id",
            client_arg="app_id",
        )
        c._handle = (
            await wa.api.upload_file(
                upload_session_id=(
                    await wa.api.create_upload_session(
                        app_id=app_id,
                        file_name=filename
                        or _template_header_formats_filename.get(
                            c.format, "pywa-template-header"
                        ),
                        file_length=len(raw_bytes),
                        file_type=mime_type
                        or _template_header_formats_default_mime_types.get(
                            c.format, "application/octet-stream"
                        ),
                    )
                )["id"],
                file=raw_bytes,
                file_offset=0,
            )
        )["h"]
    except Exception as e:
        e.add_note(
            f"Failed to upload media for component {c.__class__.__name__} with example: {example}"
        )
        raise


async def upload_template_media_params(
    *,
    wa: WhatsApp,
    sender: str,
    params: list[TemplateBaseComponent.Params | dict],
) -> None:
    """
    Internal method to upload media parameters when sending a template message.
    """
    not_uploaded = []
    for param in params:
        if isinstance(param, _BaseMediaParams) and param._resolved_media is None:
            not_uploaded.append(param)
        elif isinstance(param, Carousel.Params):
            for card_params in param.cards:
                for p in card_params.params:
                    if isinstance(p, _BaseMediaParams) and p._resolved_media is None:
                        not_uploaded.append(p)

    if not not_uploaded:
        return

    await _run_all_and_cancel_on_exception(
        *[
            _upload_param(
                wa=wa,
                sender=sender,
                p=p,
            )
            for p in not_uploaded
        ]
    )


async def _upload_param(*, wa: WhatsApp, sender: str, p: _BaseMediaParams) -> None:
    try:
        p._is_url, p._resolved_media = await resolve_media_param(
            wa=wa,
            media=p.media,
            mime_type=None,
            filename=None,
            media_type=p.format.value,
            phone_id=sender,
        )
    except Exception as e:
        e.add_note(f"Failed to upload media parameter with media: {p.media}")
        raise e



from __future__ import annotations

__all__ = [
    "ListenerTimeout",
    "ListenerCanceled",
    "ListenerStopped",
    "UserUpdateListenerIdentifier",
    "TemplateUpdateListenerIdentifier",
]

import dataclasses
import threading
from typing import TYPE_CHECKING

from . import utils

if TYPE_CHECKING:
    from .client import WhatsApp
    from .types.base_update import BaseUpdate
    from .filters import Filter


@dataclasses.dataclass(frozen=True, kw_only=True)
class BaseListenerIdentifier: ...


@dataclasses.dataclass(frozen=True, kw_only=True)
class UserUpdateListenerIdentifier(BaseListenerIdentifier):
    sender: str
    recipient: str


@dataclasses.dataclass(frozen=True, kw_only=True)
class TemplateUpdateListenerIdentifier(BaseListenerIdentifier):
    waba_id: str
    template_id: str


class ListenerTimeout(Exception):
    """
    The listener timed out

    Example:

        .. code-block:: python

                try:
                    wa.listen(..., timeout=10)
                except ListenerTimeout as e:
                    wa.send_message(to="123456", text="You took too long to respond")

    Attributes:
        timeout: The timeout that was set for the listener
    """

    def __init__(self, timeout: float):
        self.timeout = timeout

    def __str__(self):
        return f"Listener timed out after {self.timeout} seconds"


class ListenerCanceled(Exception):
    """
    The listener was canceled by a filter

    Example:

        .. code-block:: python

            try:
                wa.listen(
                    to=UserUpdateListenerIdentifier(
                        sender="123456",
                        recipient="654321"
                    ),
                    filters=filters.message & filters.text,
                    cancelers=filters.callback_button & filters.matches("cancel")
                )
            except ListenerCanceled as e:
                assert e.update.data == "cancel" # the update that caused the listener to be canceled
                wa.send_message("123456", "You cancelled the listener by clicking the `cancel` button")

    Attributes:
        update: The update that caused the listener to be canceled
    """

    def __init__(self, update: BaseUpdate | None = None):
        self.update = update

    def __str__(self):
        return (
            f"Listener was canceled by update: {self.update}"
            if self.update
            else "Listener was canceled"
        )


class ListenerStopped(Exception):
    """
    The listener was stopped manually by wa.stop_listening(...)

    Example:

        .. code-block:: python

            try:
                wa.listen(...)
            except ListenerStopped as e:
                print(e.reason) # print the reason the listener was stopped
                wa.send_message("123456", "The listener was stopped")

    Attributes:
        reason: The reason the listener was stopped (set by `wa.stop_listening(reason="...")`)
    """

    def __init__(self, reason: str | None = None):
        self.reason = reason

    def __str__(self):
        return (
            f"Listener was stopped: {self.reason}"
            if self.reason
            else "Listener was stopped"
        )


class Listener:
    def __init__(
        self,
        filters: Filter,
        cancelers: Filter,
    ):
        self.filters = filters
        self.cancelers = cancelers
        self.event: threading.Event = threading.Event()
        self.result: BaseUpdate | None = None
        self.exception: Exception | None = None

    def set_result(self, result: BaseUpdate) -> None:
        self.result = result
        self.event.set()

    def set_exception(self, exception: Exception) -> None:
        self.exception = exception
        self.event.set()

    def cancel(self, update: BaseUpdate | None = None) -> None:
        self.set_exception(ListenerCanceled(update))

    def stop(self, reason: str | None = None) -> None:
        self.set_exception(ListenerStopped(reason))

    def is_set(self) -> bool:
        return self.event.is_set()

    def apply_filters(self, wa: WhatsApp, update: BaseUpdate) -> bool:
        return not self.filters or self.filters.check_sync(wa, update)

    def apply_cancelers(self, wa: WhatsApp, update: BaseUpdate) -> bool:
        return self.cancelers and self.cancelers.check_sync(wa, update)


class _Listeners:
    def listen(
        self: WhatsApp,
        to: BaseListenerIdentifier,
        *,
        filters: Filter = None,
        cancelers: Filter = None,
        timeout: float | None = None,
    ) -> BaseUpdate:
        """
        Listen to an update.

        - You can use one of the shortcuts to listen to a specific update type:
            - :meth:`~pywa.types.sent_message.SentMessage.wait_for_reply`
            - :meth:`~pywa.types.sent_message.SentMessage.wait_for_click`
            - :meth:`~pywa.types.sent_message.SentMessage.wait_for_selection`
            - :meth:`~pywa.types.sent_message.SentMessage.wait_for_completion`
            - :meth:`~pywa.types.sent_message.SentMessage.wait_until_read`
            - :meth:`~pywa.types.sent_message.SentMessage.wait_until_delivered`

        Example:

            .. code-block:: python

                    try:
                        wa.send_message(
                            to="123456",
                            text="Send me a message",
                            buttons=[Button(title="Cancel", callback_data="cancel")]
                        )
                        update: Message = wa.listen(
                            to=UserUpdateListenerIdentifier(sender="123456", recipient="654321"),
                            filters=filters.message & filters.text,
                            cancelers=filters.callback_button & filters.matches("cancel"),
                            timeout=10
                        )
                        print(update)
                    except ListenerTimeout:
                        print("Listener timed out")
                    except ListenerCanceled:
                        print("Listener was canceled")
                    except ListenerStopped:
                        print("Listener was stopped")

        Args:
            to: The identifier of the update to listen to.
            filters: The filters to apply to the update, return the update if the filters pass.
            cancelers: The filters to cancel the listening, raise :class:`ListenerCanceled` if the update matches.
            timeout: The time to wait for the update, raise :class:`ListenerTimeout` if the time passes

        Returns:
            The update that passed the filters

        Raises:
            ListenerTimeout: If the listener timed out
            ListenerCanceled: If the listener was canceled by a filter
            ListenerStopped: If the listener was stopped manually
        """
        if self._server is utils.MISSING:
            raise ValueError(
                "You must initialize the WhatsApp client with an web app"
                " (Flask or FastAPI or custom server by setting `server` to None) in order to listen to incoming updates."
            )
        listener = Listener(
            filters=filters,
            cancelers=cancelers,
        )
        self._listeners[to] = listener
        try:
            if not listener.event.wait(timeout):
                raise ListenerTimeout(timeout)

            if listener.exception:
                raise listener.exception

            return listener.result
        finally:
            self._remove_listener(identifier=to)

    def stop_listening(
        self: WhatsApp,
        to: BaseListenerIdentifier,
        *,
        reason: str | None = None,
    ) -> None:
        """
        Stop listening to updates for a specific listener

            - Raising :class:`ListenerStopped` to the listener

        Args:
            to: The identifier of the listener to stop
            reason: The reason to stop listening

        Raises:
            ValueError: If the listener does not exist
        """

        try:
            listener = self._listeners[to]
        except KeyError:
            raise ValueError("Listener does not exist")
        listener.stop(reason)
        self._remove_listener(identifier=to)

    def _remove_listener(self: WhatsApp, identifier: BaseListenerIdentifier) -> None:
        try:
            del self._listeners[identifier]
        except KeyError:
            pass



import asyncio

import pytest
import threading
import time

from pywa import WhatsApp as WhatsAppSync, filters
from pywa_async import WhatsApp as WhatsAppAsync
from pywa.listeners import (
    ListenerTimeout,
    ListenerCanceled,
    ListenerStopped,
    UserUpdateListenerIdentifier,
)


class DummyUpdate:
    @property
    def listener_identifier(self):
        return UserUpdateListenerIdentifier(sender="123456789", recipient="987654321")


@pytest.fixture
def wa_sync():
    return WhatsAppSync(server=None, verify_token="xyz")


@pytest.fixture
def wa_async():
    return WhatsAppAsync(server=None, verify_token="xyz")


def test_listener_success_sync(wa_sync: WhatsAppSync):
    identifier = DummyUpdate().listener_identifier

    def emit_update():
        time.sleep(0.1)
        update = DummyUpdate()
        wa_sync._process_listener(update)

    threading.Thread(target=emit_update).start()
    result = wa_sync.listen(
        to=identifier, filters=filters.true, cancelers=filters.false, timeout=1
    )
    assert isinstance(result, DummyUpdate)


@pytest.mark.asyncio
async def test_listener_success_async(wa_async: WhatsAppAsync):
    identifier = DummyUpdate().listener_identifier

    async def emit_update():
        await asyncio.sleep(0.1)
        update = DummyUpdate()
        await wa_async._process_listener(update)

    asyncio.create_task(emit_update())
    result = await wa_async.listen(
        to=identifier, filters=filters.true, cancelers=filters.false, timeout=1
    )
    assert isinstance(result, DummyUpdate)


def test_listener_timeout_sync(wa_sync: WhatsAppSync):
    identifier = DummyUpdate().listener_identifier
    with pytest.raises(ListenerTimeout):
        wa_sync.listen(
            to=identifier, filters=filters.true, cancelers=filters.false, timeout=0.1
        )


@pytest.mark.asyncio
async def test_listener_timeout_async(wa_async: WhatsAppAsync):
    identifier = DummyUpdate().listener_identifier
    with pytest.raises(ListenerTimeout):
        await wa_async.listen(
            to=identifier, filters=filters.true, cancelers=filters.false, timeout=0.1
        )


def test_listener_canceled_sync(wa_sync: WhatsAppSync):
    identifier = DummyUpdate().listener_identifier

    def emit_update():
        time.sleep(0.1)
        update = DummyUpdate()
        wa_sync._process_listener(update)

    threading.Thread(target=emit_update).start()
    with pytest.raises(ListenerCanceled):
        wa_sync.listen(
            to=identifier, filters=filters.false, cancelers=filters.true, timeout=0.2
        )


@pytest.mark.asyncio
async def test_listener_canceled_async(wa_async: WhatsAppAsync):
    identifier = DummyUpdate().listener_identifier

    async def emit_update():
        await asyncio.sleep(0.1)
        update = DummyUpdate()
        await wa_async._process_listener(update)

    asyncio.create_task(emit_update())
    with pytest.raises(ListenerCanceled):
        await wa_async.listen(
            to=identifier, filters=filters.false, cancelers=filters.true, timeout=0.2
        )


def test_listener_stopped_sync(wa_sync: WhatsAppSync):
    identifier = DummyUpdate().listener_identifier

    def stop_listener():
        time.sleep(0.1)
        wa_sync.stop_listening(to=identifier, reason="manual")

    threading.Thread(target=stop_listener).start()
    with pytest.raises(ListenerStopped) as exc_info:
        wa_sync.listen(
            to=identifier, filters=filters.true, cancelers=filters.false, timeout=1
        )
    assert exc_info.value.reason == "manual"


@pytest.mark.asyncio
async def test_listener_stopped_async(wa_async: WhatsAppAsync):
    identifier = DummyUpdate().listener_identifier

    async def stop_listener():
        await asyncio.sleep(0.1)
        wa_async.stop_listening(to=identifier, reason="manual")

    asyncio.create_task(stop_listener())
    with pytest.raises(ListenerStopped) as exc_info:
        await wa_async.listen(
            to=identifier, filters=filters.true, cancelers=filters.false, timeout=1
        )
    assert exc_info.value.reason == "manual"



"""This module contains the types related to media."""

from __future__ import annotations

__all__ = [
    "Image",
    "Video",
    "Sticker",
    "Document",
    "Audio",
    "MediaUrlResponse",
]

from pywa.types.others import SuccessResult
from .. import utils
from pywa.types.media import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.media import (
    Image as _Image,
    Video as _Video,
    Sticker as _Sticker,
    Document as _Document,
    Audio as _Audio,
    MediaUrlResponse as _MediaUrlResponse,
)  # noqa MUST BE IMPORTED FIRST

import dataclasses
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..client import WhatsApp


class Media:
    """Base class for all media types."""

    def __init__(self, _client: WhatsApp, id: str):
        self._client = _client
        self.id = id

    async def get_media_url(self) -> str:
        """Gets the URL of the media. (expires after 5 minutes)"""
        return (await self._client.get_media_url(media_id=self.id)).url

    async def download(
        self,
        path: str | None = None,
        filename: str | None = None,
        in_memory: bool = False,
        **kwargs,
    ) -> bytes | str:
        """
        Download a media file from WhatsApp servers.
            - Same as :func:`~pywa.client.WhatsApp.download_media` with ``media_url=media.get_media_url()``

        >>> message.image.download()

        Args:
            path: The path where to save the file (if not provided, the current working directory will be used).
            filename: The name of the file (if not provided, it will be guessed from the URL + extension).
            in_memory: Whether to return the file as bytes instead of saving it to disk (default: False).
            **kwargs: Additional arguments to pass to ``httpx.get(...)``.

        Returns:
            The path of the saved file if ``in_memory`` is False, the file as bytes otherwise.
        """
        return await self._client.download_media(
            url=(await self.get_media_url())
            if not hasattr(self, "url")
            else self.url,  # MediaUrlResponse
            path=path,
            filename=filename,
            in_memory=in_memory,
            **kwargs,
        )

    async def delete(
        self, *, phone_id: str | int | None = utils.MISSING
    ) -> SuccessResult:
        """
        Deletes the media from WhatsApp servers.

        Args:
            phone_id: The phone ID to delete the media from (optional, If included, the operation will only be processed if the ID matches the ID of the business phone number that the media was uploaded on. pass None to use the client's phone ID).
        """
        return await self._client.delete_media(media_id=self.id, phone_id=phone_id)


class BaseUserMedia(Media):
    """Base class for all media types."""

    @classmethod
    def from_flow_completion(
        cls, client: WhatsApp, media: dict[str, str]
    ) -> BaseUserMedia:
        """
        Create a media object from the media dict returned by the flow completion.

        - You can use the shortcut :meth:`~pywa_async.types.FlowCompletion.get_media`

        Example:
            >>> from pywa_async import WhatsApp, types
            >>> wa = WhatsApp(...)
            >>> @wa.on_flow_completion
            ... async def on_flow_completion(_: WhatsApp, flow: types.FlowCompletion):
            ...     img = types.Image.from_flow_completion(client=wa, media=flow.response['media'])
            ...     await img.download()

        Args:
            client: The WhatsApp client.
            media: The media dict returned by the flow completion.

        Returns:
            The media object (Image, Video, Sticker, Document, Audio).
        """
        # noinspection PyUnresolvedReferences
        return cls.from_dict(media, _client=client)


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Image(BaseUserMedia, _Image):
    """
    Represents an received image.

    Attributes:
        id: The ID of the file (can be used to download or re-send the image).
        sha256: The SHA256 hash of the image.
        mime_type: The MIME type of the image.
    """


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Video(BaseUserMedia, _Video):
    """
    Represents a video.

    Attributes:
        id: The ID of the file (can be used to download or re-send the video).
        sha256: The SHA256 hash of the video.
        mime_type: The MIME type of the video.
    """


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Sticker(BaseUserMedia, _Sticker):
    """
    Represents a sticker.

    Attributes:
        id: The ID of the file (can be used to download or re-send the sticker).
        sha256: The SHA256 hash of the sticker.
        mime_type: The MIME type of the sticker.
        animated: Whether the sticker is animated.
    """


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Document(BaseUserMedia, _Document):
    """
    Represents a document.

    Attributes:
        id: The ID of the file (can be used to download or re-send the document).
        sha256: The SHA256 hash of the document.
        mime_type: The MIME type of the document.
        filename: The filename of the document (optional).
    """


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Audio(BaseUserMedia, _Audio):
    """
    Represents an audio.

    Attributes:
        id: The ID of the file (can be used to download or re-send the audio).
        sha256: The SHA256 hash of the audio.
        mime_type: The MIME type of the audio.
        voice: Whether the audio is a voice message or just an audio file.
    """


@dataclasses.dataclass(frozen=True, slots=True)
class MediaUrlResponse(Media, _MediaUrlResponse):
    """
    Represents a media response.

    Attributes:
        id: The ID of the media.
        url: The URL of the media (valid for 5 minutes).
        mime_type: The MIME type of the media.
        sha256: The SHA256 hash of the media.
        file_size: The size of the media in bytes.
    """
